"""
FIXME module add docstrings
"""
from unittest import TestCase, main
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from easy_gscv.models import GSCV

# pylint: disable=C0103
# 'x' and 'y' are very descriptive in the realm of datascience.

# pylint: disable=W0212
# Method call required for testing


class TestProperties(TestCase):
    """
    Make all model properties work as they should.
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.clf = KNeighborsClassifier()
        self.model = GSCV(self.clf, self.x, self.y)
        self.model_dict = {
            'KNeighborsClassifier': KNeighborsClassifier(),
            'RandomForestClassifier': RandomForestClassifier(),
            'GradientBoostingClassifier': GradientBoostingClassifier(),
            'MLPClassifier': MLPClassifier(),
            'LogisticRegression': LogisticRegression(),
        }

    def test_classifiers(self):
        """
        Test that the 'classifiers' property returns
        a list of valid classifiers
        """
        self.assertEqual(
            self.model.classifiers, self.model_dict.keys())


class TestCLFTypes(TestCase):
    """
    Test that the clf argument can handle multiple types.
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']

    def test_sklearn_classifier(self):
        """
        Test that object accepts a sklearn classifier
        """
        clf = KNeighborsClassifier()
        model = GSCV(clf, self.x, self.y)
        self.assertEqual(
            type(model._get_model(clf)), type(KNeighborsClassifier())
        )

    def test_string(self):
        """
        Test that object accepts a string
        """
        clf = 'KNeighborsClassifier'
        model = GSCV(clf, self.x, self.y)
        self.assertEqual(
            type(model._get_model(clf)), type(KNeighborsClassifier())
        )


class TestExceptions(TestCase):
    """
    Make sure that the exceptions trigger when they should.
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']

        class Nothing:
            """Nonsense object that should not pass"""
            def __init__(self, a):
                self.a = a

        self.wrong_object = Nothing(a='horse')
        self.wrong_clf = DecisionTreeClassifier()
        self.valid_clf = KNeighborsClassifier()

    def test_check_model_not_a_model(self):
        """
        Make sure that using a type that is not a model
        raises the correct error message.
        """
        with self.assertRaises(ValueError):
            GSCV('doesnotexist', self.x, self.y)

    def test_check_model_not_a_sklearn_model(self):
        """
        Make sure that using a non-sklearn model raises
        the correct error message.
        """
        with self.assertRaises(TypeError):
            GSCV(self.wrong_object, self.x, self.y)

    def test_check_wrong_scikit_model(self):
        """
        Make sure that using a non-valid sklearn classifier
        raises the correct error message
        """
        with self.assertRaises(ValueError):
            GSCV(self.wrong_clf, self.x, self.y)


class TestKNeighborsClassifier(TestCase):
    """
    Make sure that the model is created and the methods work
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.valid_clf = KNeighborsClassifier()

    def test_default_params(self):
        """
        Make sure that the correct default parameters
        are selected for the model
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        self.assertEqual(
            model.params, {
                'n_neighbors': [3, 5, 8, 10, 15],
                'weights': ['uniform', 'distance'],
            }
        )

    def test_custom_params(self):
        """Test that custom params override the default ones"""
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'n_neighbors': [3, 15],
            'weights': ['uniform'],
        })
        self.assertEqual(
            model.params, {
                'n_neighbors': [3, 15],
                'weights': ['uniform'],
            }
        )

    def test_create(self):
        """Test that the create method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y)
        result = model.create()
        self.assertTrue(result is not None)

    def test_score(self):
        """Test that the create method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y)
        score = model.score()
        self.assertTrue(score is not None)
        self.assertTrue(0 <= score <= 1)

    def test_get_best_estimator(self):
        """Test that the 'get_best_estimator' method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y)
        best_model = model.get_best_estimator()
        self.assertTrue(best_model is not None)

    def test_get_fit_details(self):
        """Test that the 'get_fit_details' method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y)
        fit_details = model.get_fit_details()
        self.assertTrue(fit_details is not None)


class TestLogisticRegression(TestCase):
    """ Make sure that the model is created and the methods work """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.valid_clf = LogisticRegression()

    def test_default_params(self):
        """
        Make sure that the correct default parameters
        are selected for the model
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        self.assertEqual(
            model.params, {
                'C': [0.01, 1, 100],
                'penalty': ['l1', 'l2']
            }
        )

    def test_custom_params(self):
        """Test that custom params override the default ones"""
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'C': [1, 100],
            'penalty': ['l2']
        })
        self.assertEqual(
            model.params, {
                'C': [1, 100],
                'penalty': ['l2']
            }
        )

    def test_create(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        result = model.create()
        self.assertTrue(result is not None)

    def test_score(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        score = model.score()
        self.assertTrue(score is not None)
        self.assertTrue(0 <= score <= 1)

    def test_get_best_estimator(self):
        """
        Test that the 'get_best_estimator' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        best_model = model.get_best_estimator()
        self.assertTrue(best_model is not None)

    def test_get_fit_details(self):
        """
        Test that the 'get_fit_details' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        fit_details = model.get_fit_details()
        self.assertTrue(fit_details is not None)


class TestMLPClassifier(TestCase):
    """
    Make sure that the model is created and the methods work
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.valid_clf = MLPClassifier()

    def test_default_params(self):
        """
        Make sure that the correct default parameters
        are selected for the model
        """
        model = GSCV(self.valid_clf, self.x, self.y)
        self.assertEqual(
            model.params, {
                'hidden_layer_sizes': [
                    (10,),
                    (10, 10),
                    (10, 10, 10)
                ],
                'alpha': [0.0001, 0.01, 0.1, 1],
                'solver': ['lbfgs'],
            }
        )

    def test_custom_params(self):
        """Test that custom params override the default ones"""
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'alpha': [0.0001, 0.01, 0.1, 1]
        })
        self.assertEqual(
            model.params, {
                'alpha': [0.0001, 0.01, 0.1, 1]
            }
        )

    def test_create(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'alpha': [0.0001, 0.01, 0.1, 1]
        })
        result = model.create()
        self.assertTrue(result is not None)

    def test_score(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'alpha': [0.0001, 0.01, 0.1, 1]
        })
        score = model.score()
        self.assertTrue(score is not None)
        self.assertTrue(0 <= score <= 1)

    def test_get_best_estimator(self):
        """
        Test that the 'get_best_estimator' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'alpha': [0.0001, 0.01, 0.1, 1]
        })
        best_model = model.get_best_estimator()
        self.assertTrue(best_model is not None)

    def test_get_fit_details(self):
        """
        Test that the 'get_fit_details' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, params={
            'alpha': [0.0001, 0.01, 0.1, 1]
        })
        fit_details = model.get_fit_details()
        self.assertTrue(fit_details is not None)


class TestRandomForestClassifier(TestCase):
    """
    Make sure that the model is created and the methods work
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.valid_clf = RandomForestClassifier()

    # def test_default_params(self):
    #     model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1)
    #     self.assertEqual(
    #         model.params, {
    #             'n_estimators': [100, 500, 1000],
    #             'max_features': ['sqrt', 'log2', None],
    #             'max_depth': [None, 3, 5],
    #         }
    #     )

    def test_custom_params(self):
        """Test that custom params override the default ones"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'max_features': ['sqrt'],
            'max_depth': [3],
        })
        self.assertEqual(
            model.params, {
                'n_estimators': [100],
                'max_features': ['sqrt'],
                'max_depth': [3],
            }
        )

    def test_create(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'max_features': ['sqrt'],
            'max_depth': [3],
        })
        result = model.create()
        self.assertTrue(result is not None)

    def test_score(self):
        """
        Test that the create method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'max_features': ['sqrt'],
            'max_depth': [3],
        })
        score = model.score()
        self.assertTrue(score is not None)
        self.assertTrue(0 <= score <= 1)

    def test_get_best_estimator(self):
        """
        Test that the 'get_best_estimator' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'max_features': ['sqrt'],
            'max_depth': [3],
        })
        best_model = model.get_best_estimator()
        self.assertTrue(best_model is not None)

    def test_get_fit_details(self):
        """
        Test that the 'get_fit_details' method returns a value
        """
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'max_features': ['sqrt'],
            'max_depth': [3],
        })
        fit_details = model.get_fit_details()
        self.assertTrue(fit_details is not None)


class TestGradientBoostingClassifier(TestCase):
    """
    Make sure that the model is created and the methods work
    """

    def setUp(self):
        # Get data
        from sklearn import datasets
        iris = datasets.load_iris()
        self.x = iris['data']
        self.y = iris['target']
        self.valid_clf = GradientBoostingClassifier()

    # def test_default_params(self):
    #     model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1)
    #     self.assertEqual(
    #         model.params, {
    #             'n_estimators': [100, 500, 1000],
    #             'learning_rate': [0.1, 0.5, 1],
    #             'max_depth': [1, 3, 5],
    #             'max_features': ['sqrt', 'log2', None],
    #         }
    #     )

    def test_custom_params(self):
        """Test that custom params override the default ones"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'learning_rate': [0.1, 0.5],
            'max_depth': [1, 3],
            'max_features': ['sqrt'],
        })
        self.assertEqual(
            model.params, {
                'n_estimators': [100],
                'learning_rate': [0.1, 0.5],
                'max_depth': [1, 3],
                'max_features': ['sqrt'],
            }
        )

    def test_create(self):
        """Test that the create method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'learning_rate': [0.1],
            'max_depth': [1],
            'max_features': ['sqrt'],
        })
        result = model.create()
        self.assertTrue(result is not None)

    def test_score(self):
        """Test that the create method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'learning_rate': [0.1],
            'max_depth': [1],
            'max_features': ['sqrt'],
        })
        score = model.score()
        self.assertTrue(score is not None)
        self.assertTrue(0 <= score <= 1)

    def test_get_best_estimator(self):
        """Test that the 'get_best_estimator' method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'learning_rate': [0.1],
            'max_depth': [1],
            'max_features': ['sqrt'],
        })
        best_model = model.get_best_estimator()
        self.assertTrue(best_model is not None)

    def test_get_fit_details(self):
        """Test that the 'get_fit_details' method returns a value"""
        model = GSCV(self.valid_clf, self.x, self.y, n_jobs=-1, params={
            'n_estimators': [100],
            'learning_rate': [0.1],
            'max_depth': [1],
            'max_features': ['sqrt'],
        })
        fit_details = model.get_fit_details()
        self.assertTrue(fit_details is not None)


if __name__ == '__main__':
    print('\n\n')
    print("Running tests!")
    print(
        'Warning: Despite the use of a (relatively) test small dataset '
        'these tests can take up to 2 minutes to complete on a relativly '
        'modern computer due to the number of models being trained.'
        'The models have been set up to utilize all available cpu cores '
        'to speed up the process and the more computationally models have '
        'their "test_default_params" method tests disabled, while their '
        'other tests use lighter-than-default params.'
        '\n\nResetting these values to their default parameters '
        'will increase the runtime significantly'
        )
    print('\n\n')
    main()



"""
High level objects to automate large parts of the classifier training workflow.
"""
from typing import Optional, Dict, KeysView, Union
from sklearn.model_selection import (  # type: ignore
    train_test_split, GridSearchCV
)  # type: ignore
from sklearn.ensemble import (  # type: ignore
    RandomForestClassifier, GradientBoostingClassifier
    )  # type: ignore
from sklearn.neighbors import KNeighborsClassifier  # type: ignore
from sklearn.linear_model import LogisticRegression  # type: ignore
from sklearn.neural_network import MLPClassifier  # type: ignore


# Create sklearn_classifiers Union object for mypy
ScikitClassifiers = Union[
    RandomForestClassifier, GradientBoostingClassifier,
    KNeighborsClassifier, LogisticRegression,
    MLPClassifier
]

# pylint: disable=too-many-instance-attributes
# pylint: disable-msg=too-many-arguments
# Arguments are needed to build the models.


class GSCV:
    """
    This object will build trained classifiers with the specified parameters.
    Training / testing datasets are created automatically from the
    [X,y] keywords and the models are trained using
    grid search and cross validation.
    """

    # Dictionary with all suported classifiers as values
    # and matching strings as keys
    model_dict = {
        'KNeighborsClassifier': KNeighborsClassifier(),
        'RandomForestClassifier': RandomForestClassifier(),
        'GradientBoostingClassifier': GradientBoostingClassifier(),
        'MLPClassifier': MLPClassifier(),
        'LogisticRegression': LogisticRegression(),
    }

    def __init__(
            self, clf, X, y, cross_vals: int = 10, random_state: int = 42,
            test_size: float = 0.33, n_jobs: int = 1,
            params: Optional[Dict[str, Union[str, float]]] = None
    ) -> None:

        # Check if classifier is valid and assign it if true
        self.clf = self._get_model(clf)

        # Split data into Train and test sets
        x_train, x_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=random_state
        )

        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test
        self.cross_vals = cross_vals
        self.random_state = random_state
        self.n_jobs = n_jobs

        # Set params
        if params is None:
            params = self._get_model_params()
        self.params = params

        # Create model on initialization
        self.gs_model = self.create()

    @staticmethod
    def _get_model_name(clf, target='classifier') -> str:
        """
        With sklearn classifiers missing the '__name__' attribute,
        getting their name has proven to be suprisingly tricky.
        """
        if target == 'classifier':
            name = str(
                clf.__class__
            ).replace('>', '').split('.')[-1].replace("'", "")
        if target == 'class':
            name = str(clf.__class__).replace(
                "class '", ''
            ).replace('<', '').split('.')[0]
        return name

    def _get_model(
            self, clf: Union[str, ScikitClassifiers]
    ) -> ScikitClassifiers:
        """
        Fetches to correct model for the provided clf string
        or returns an error message if the clf is invalid.
        """

        error_message = """
        Model not supported! Please use one of the following classifiers:
        \n {}
        """.format(", ".join(map(str, self.model_dict)))

        # If a string is passed instead of a sklearn classifier
        # the program should check if the string matches a supported
        # classifier
        if isinstance(clf, str):
            if clf not in self.model_dict.keys():
                raise ValueError(error_message)
            return self.model_dict[clf]

        if self._get_model_name(clf, 'class') != 'sklearn':
            raise TypeError(
                'The classifier is not a valid scikit-learn classifier!'
            )

        if self._get_model_name(clf, 'classifier') \
                not in self.model_dict.keys():
            raise ValueError(error_message)

        name = self._get_model_name(clf, 'classifier')
        return self.model_dict[name]

    def _get_model_params(self) -> Dict:
        """
        An Internal method to get a matching model_param dict for the
        provided classifier, called only if no params are provided.
        """
        # Get number of features
        if self.x_train.shape[1] > 10:
            no_ftrs = self.x_train.shape[1]
        else:
            no_ftrs = 10

        # Choice of algorithm depends on size of dataset
        if self.x_train.shape[1] < 500:
            solver = 'lbfgs'
        else:
            solver = 'adam'

        param_dict = {
            'KNeighborsClassifier': {
                'n_neighbors': [3, 5, 8, 10, 15],
                'weights': ['uniform', 'distance'],
            },
            'RandomForestClassifier': {
                'n_estimators': [100, 500, 1000],
                'max_features': ['sqrt', 'log2', None],
                'max_depth': [None, 3, 5],
            },
            'GradientBoostingClassifier': {
                'n_estimators': [100, 500, 1000],
                'learning_rate': [0.1, 0.5, 1],
                'max_depth': [1, 3, 5],
                'max_features': ['sqrt', 'log2', None],
            },
            'MLPClassifier': {
                'hidden_layer_sizes': [
                    (no_ftrs,),
                    (no_ftrs, no_ftrs),
                    (no_ftrs, no_ftrs, no_ftrs)
                ],
                'alpha': [0.0001, 0.01, 0.1, 1],
                'solver': [solver],
            },
            'LogisticRegression': {
                'C': [0.01, 1, 100],
                'penalty': ['l1', 'l2']
            }
        }
        params = param_dict[self._get_model_name(self.clf, 'classifier')]
        return params

    @property
    def classifiers(self) -> KeysView[str]:
        """Returns a list of all valid classifiers"""
        return self.model_dict.keys()

    def create(self) -> GridSearchCV:
        """
        Train the classifier with gridsearch and cross evaluation enabled.
        Return GSCV model. Warning: The returned model is not a classifier!
        """
        gs_model = GridSearchCV(
            self.clf, self.params, n_jobs=self.n_jobs, cv=self.cross_vals
        )
        gs_model.fit(self.x_train, self.y_train)
        return gs_model

    def score(self) -> float:
        """
        Scores the best_estimator on the X_test and y_test
        datasets and returns it.
        """
        return self.gs_model.best_estimator_.score(self.x_test, self.y_test)

    def get_best_estimator(self) -> object:
        """
        Return the best_estimator object
        """
        return self.gs_model.best_estimator_

    def get_fit_details(self) -> str:
        """
        Print a table that shows the (training) scores
        for the various parameter combinations.
        """
        table = ""
        means = self.gs_model.cv_results_['mean_test_score']
        stds = self.gs_model.cv_results_['std_test_score']
        for mean, std, params in zip(
                means, stds, self.gs_model.cv_results_['params']
        ):
            table += "%0.3f (+/-%0.03f) for %r" % (mean, std * 2, params)
        return table



"""
Script for packaging.
"""
import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="easy-gscv",
    version="0.1.0",
    author="Jasper Koops",
    author_email="jasperkoops91@gmail.com",
    description=(
        "A high level library gridsearch / cross evaluation library "
        "for scikit-learn"
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/Jasper-Koops/easy-gscv",
    packages=setuptools.find_packages(exclude=['venv']),
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
    ],
    classifiers=(
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)
